# Copyright (c) Qotto, 2021

"""
Record abstract base class
"""

from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, Optional
from uuid import UUID, uuid4

from semver import VersionInfo

from eventy.record.errors import RecordAttributeTypeError, RecordAttributeValueError
from eventy.trace_id import correlation_id_var
from eventy.trace_id.generator import gen_trace_id

__all__ = [
    'Record',
    'RecordType',
]


class RecordType(Enum):
    @classmethod
    def list(cls):
        return [type.value for type in cls]

    EVENT = 'EVENT'
    REQUEST = 'REQUEST'
    RESPONSE = 'RESPONSE'


class Record:
    """
    Common Record interface

    This class is a base for Event, Request, and Response. It should not be instantiated.
    """

    def __init__(
        self,
        protocol_version: str,
        namespace: str,
        name: str,
        version: str,
        source: str,
        uuid: Optional[str] = None,
        correlation_id: Optional[str] = None,
        partition_key: Optional[str] = None,
        date: Optional[datetime] = None,
        context: Optional[Dict[str, Dict[str, Any]]] = None,
        data: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Initialize Record attributes

        :param protocol_version: Version of the Eventy protocol in SemVer format
        :param namespace: Namespace of the record schema in URN format
        :param name: Name of the record schema
        :param version: Version of the record schema in SemVer format
        :param source: Emitter of the record in URN format
        :param uuid: UUID v4 of the record. Default: auto-generated
        :param correlation_id: Correlation ID, propagated. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Context data dictionary, propagated. Default: None
        :param data: Domain data dictionary. Default: None

        :raises RecordAttributeError:
        """
        self.protocol_version = protocol_version
        self.namespace = namespace
        self.name = name
        self.version = version
        self.source = source
        if uuid is None:
            uuid = str(uuid4())
        self.uuid = uuid
        if correlation_id is None:
            correlation_id = correlation_id_var.get()
            if not correlation_id:
                correlation_id = gen_trace_id(self.name)
        self.correlation_id = correlation_id
        # partition_key can be None
        self.partition_key = partition_key
        if date is None:
            date = datetime.now(tz=timezone.utc)
        self.date = date
        if context is None:
            context = {}
        self.context = context
        if data is None:
            data = {}
        self.data = data

    @property
    def type(self) -> RecordType:
        """
        Type of the record: EVENT | REQUEST | RESPONSE
        """
        raise NotImplementedError

    @property
    def protocol_version(self) -> str:
        """
        Eventy protocol version (SemVer)
        """
        return self._protocol_version

    @protocol_version.setter
    def protocol_version(self, protocol_version: str) -> None:
        try:
            VersionInfo.parse(protocol_version)
        except ValueError:
            raise RecordAttributeValueError('protocol_version', protocol_version, 'Not in SemVer format')
        self._protocol_version = protocol_version

    @property
    def qualified_name(self) -> str:
        return f'{self.namespace}:{self.name}'

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def namespace(self) -> str:
        return self._namespace

    @namespace.setter
    def namespace(self, namespace: str) -> None:
        if not namespace.startswith('urn:'):
            raise RecordAttributeValueError('namespace', namespace, 'URN format should start with "urn:"')
        self._namespace = namespace

    @property
    def version(self) -> str:
        """
        Record schema version (SemVer)
        """
        return self._version

    @version.setter
    def version(self, version: str) -> None:
        try:
            VersionInfo.parse(version)
        except ValueError:
            raise RecordAttributeValueError('version', version, 'Not in SemVer format')
        self._version = version

    @property
    def source(self) -> str:
        """
        Record source (URN)
        """
        return self._source

    @source.setter
    def source(self, source: str) -> None:
        if not source.startswith('urn:'):
            raise RecordAttributeValueError('source', source, 'URN format should start with "urn:"')
        self._source = source

    @property
    def uuid(self) -> str:
        """
        Record unique identifier (UUID v4)
        """
        return self._uuid

    @uuid.setter
    def uuid(self, uuid: str):
        if uuid is None:
            uuid = str(uuid4())
        try:
            UUID(uuid)
        except TypeError as e:
            raise RecordAttributeValueError('uuid', uuid, str(e)) from e
        except ValueError as e:
            raise RecordAttributeValueError('uuid', uuid, str(e)) from e
        self._uuid = uuid

    @property
    def correlation_id(self) -> str:
        """
        Record correlation_id to be propagated
        """
        return self._correlation_id

    @correlation_id.setter
    def correlation_id(self, correlation_id: str) -> None:
        if isinstance(correlation_id, str):
            self._correlation_id = correlation_id
        else:
            raise RecordAttributeTypeError('correlation_id', str, correlation_id)

    @property
    def partition_key(self) -> Optional[str]:
        """
        Key for partitioned states, e.g. Kafka topics partitions
        """
        return self._partition_key

    @partition_key.setter
    def partition_key(self, partition_key: Optional[str]):
        if partition_key is None or isinstance(partition_key, str):
            self._partition_key = partition_key
        else:
            raise RecordAttributeTypeError('partition_key', str, partition_key)

    @property
    def date(self) -> datetime:
        """
        Date of the event (unrelated to event production)
        """
        return self._date

    @date.setter
    def date(self, date: datetime) -> None:
        """
        Set the date, keeping a millisecond precision.

        :param date: record date, as a datetime, or iso str, or timestamp int
        """
        if isinstance(date, datetime):
            ts = int(date.timestamp() * 1000) / 1000
            date = datetime.fromtimestamp(ts, timezone.utc)
            self._date = date
        else:
            raise RecordAttributeTypeError('date', datetime, date)

    @property
    def context(self) -> Dict[str, Dict[str, Any]]:
        """
        Execution context to be propagated
        """
        return self._context

    @context.setter
    def context(self, context: Dict[str, Dict[str, Any]]) -> None:
        if not isinstance(context, Dict):
            raise RecordAttributeTypeError('context', Dict, context)
        for key, val in context.items():
            if not isinstance(key, str):
                raise RecordAttributeTypeError(f'context key', str, key)
            if not isinstance(val, Dict):
                raise RecordAttributeTypeError(f'value for key {key}', 'Dict', val)
        self._context = context

    @property
    def data(self) -> Dict[str, Any]:
        """
        Actual record payload
        """
        return self._data

    @data.setter
    def data(self, data) -> None:
        if not isinstance(data, Dict):
            raise RecordAttributeTypeError('data', Dict, data)
        for key in data:
            if not isinstance(key, str):
                raise RecordAttributeTypeError('data key', 'str', key)
        self._data = data

    def __str__(self) -> str:
        return f"{self.qualified_name} ({self.type.value}) v{self.version} [CID:{self.correlation_id}] #{self.uuid} ${self.partition_key}"

    def _debug_str(self) -> str:
        """
        With data and context content. For debug purposes.
        """
        return self.__str__() + f"\n      CONTEXT={self.context}\n      DATA={self.data}"



from typing import Iterable, Optional, Union

from eventy.messaging.errors import MessagingError
from eventy.messaging.store import RecordStore, Cursor
from eventy.record import Record

__all__ = [
    'MemoryStore',
]


class _ReadTopic:
    def __init__(
        self,
        topic: str,
    ) -> None:
        self.topic = topic
        self.records: list[Record] = []
        self.read_offset = 0
        self.acked_offset = 0

    def add(
        self,
        records: Union[Record, Iterable[Record]],
    ) -> None:
        if not isinstance(records, Iterable):
            records = [records]
        self.records.extend(records)

    def read(
        self,
        max_count: int = 1,
    ) -> list[Record]:
        read_records = self.records[self.read_offset:self.read_offset + max_count]
        self.read_offset += len(read_records)
        return read_records

    def ack(self, count: int = None) -> None:
        if count is None:
            self.acked_offset = self.read_offset
        else:
            if (self.acked_offset + count) > self.read_offset:
                raise MessagingError('Cannot ack more than read.')
            self.acked_offset += count

    def reset(self):
        self.read_offset = self.acked_offset

    def clear(self) -> None:
        self.records = []
        self.read_offset = 0
        self.acked_offset = 0


class _WriteTopic:
    def __init__(
        self,
        topic: str,
    ) -> None:
        self.topic = topic
        self.records: list[Record] = []
        self.committed_offset = 0

    def write_committed(
        self,
        record: Record,
    ) -> None:
        self.records.insert(self.committed_offset, record)
        self.committed_offset += 1

    def write_uncommitted(
        self,
        record: Record,
    ) -> None:
        self.records.append(record)

    def commit(self, count: int = None) -> None:
        if count is None:
            self.committed_offset = len(self.records)
        else:
            if (self.committed_offset + count) > len(self.records):
                raise MessagingError("Cannot commit more than written.")
            self.committed_offset += count

    def reset(self) -> None:
        self.records = self.records[:self.committed_offset]

    def clear(self) -> None:
        self.records = []
        self.committed_offset = 0


class _Transaction:
    def __init__(self):
        self.reads: dict[str, int] = {}
        self.writes: dict[str, int] = {}

    def add_read(self, topic: str, count: int = 1) -> None:
        if topic not in self.reads:
            self.reads[topic] = 0
        self.reads[topic] += count

    def add_write(self, topic: str, count: int = 1) -> None:
        if topic not in self.writes:
            self.writes[topic] = 0
        self.writes[topic] += count


class MemoryStore(RecordStore):
    """
    In-memory record store, essentially for testing and debug purposes
    """

    def __init__(self) -> None:
        """
        Initialize an empty store
        """
        super().__init__()

        self.read_topics: dict[str, _ReadTopic] = {}
        self.write_topics: dict[str, _WriteTopic] = {}
        self.transaction: Optional[_Transaction] = None

    def register_topic(self, topic: str, cursor: Cursor = Cursor.ACKNOWLEDGED) -> None:
        if cursor != Cursor.ACKNOWLEDGED:
            raise MessagingError(f"Only ACKNOWLEDGED cursors are supported.")
        self.read_topics[topic] = _ReadTopic(topic)

    def read(
        self,
        max_count: int = 1,
        timeout_ms: Optional[int] = None,
        auto_ack: bool = False
    ) -> list[Record]:
        if auto_ack:
            raise MessagingError(f"Auto-ack is not supported.")
        records: list[Record] = []
        for topic, topic_read in self.read_topics.items():
            reads = topic_read.read(max_count - len(records))
            records.extend(reads)
        return records

    def ack(self, timeout_ms=None) -> None:
        for topic, topic_read in self.read_topics.items():
            if self.transaction:
                self.transaction.add_read(topic, topic_read.read_offset - topic_read.acked_offset)
            else:
                topic_read.ack()

    def write(self, record: Record, topic: str, timeout_ms=None) -> None:
        if topic not in self.write_topics:
            self.write_topics[topic] = _WriteTopic(topic)
        write_topic = self.write_topics[topic]
        if self.transaction:
            self.transaction.add_write(topic)
            write_topic.write_uncommitted(record)
        else:
            write_topic.write_committed(record)

    def write_now(self, record: Record, topic: str, timeout_ms=None) -> None:
        if topic not in self.write_topics:
            self.write_topics[topic] = _WriteTopic(topic)
        write_topic = self.write_topics[topic]
        write_topic.write_committed(record)

    def start_transaction(self) -> None:
        if self.transaction:
            raise MessagingError(f"Already in a transaction.")
        self.transaction = _Transaction()

    def commit(self, timeout_ms: Optional[int] = None) -> None:
        if not self.transaction:
            raise MessagingError(f"Not in a transaction.")
        for topic, read_count in self.transaction.reads.items():
            self.read_topics[topic].ack(read_count)
        for topic, write_count in self.transaction.writes.items():
            self.write_topics[topic].commit(write_count)
        self.transaction = None

    def abort(self, timeout_ms: Optional[int] = None) -> None:
        if not self.transaction:
            raise MessagingError(f"Not in a transaction.")
        for topic, read_count in self.transaction.reads.items():
            self.read_topics[topic].reset()
        for topic, write_count in self.transaction.writes.items():
            self.write_topics[topic].reset()
        self.transaction = None

    def add_record(self, records: Union[Record, Iterable[Record]], topic: str) -> None:
        """
        Add one or multiple records to be read.

        :param records: record(s) to be added
        :param topic: topic the record(s) should be added to
        """
        if topic not in self.read_topics:
            raise MessagingError(f"Topic {topic} not registered.")
        self.read_topics[topic].add(records)

    def get_committed_records(self, topic: str = None) -> list[Record]:
        """
        Get records written and committed to the store.

        :param topic: Optionally specify a topic to get records for
        :return: List of Records
        """
        if topic:
            topics = [topic]
        else:
            topics = list(self.write_topics.keys())
        records: list[Record] = []
        for topic in topics:
            if topic in self.write_topics:
                records.extend(self.write_topics[topic].records[:self.write_topics[topic].committed_offset])
        return records

    def get_acked_records(self, topic: str = None) -> list[Record]:
        """
        Get records written and acked to the store.

        :param topic: Optionally specify a topic to get records for
        :return: List of Records
        """
        if topic:
            topics = [topic]
        else:
            topics = list(self.read_topics.keys())
        records: list[Record] = []
        for topic in topics:
            if topic in self.read_topics:
                records.extend(self.read_topics[topic].records[:self.read_topics[topic].acked_offset])
        return records

    def clear(self) -> None:
        """
        Completely clear the store (remove all data).
        """
        for write_topic in self.write_topics.values():
            write_topic.clear()
        for read_topic in self.read_topics.values():
            read_topic.clear()
        self.transaction = None

    def reset(self) -> None:
        """
        Reset the store to the last committed state.
        """
        for write_topic in self.write_topics.values():
            write_topic.reset()
        for read_topic in self.read_topics.values():
            read_topic.reset()
        self.transaction = None



import logging
import time
from typing import Optional, Iterable

from eventy.messaging.agent import Agent, Handler, Guarantee
from eventy.messaging.service import Service
from eventy.messaging.store import Cursor, RecordStore, RecordWriteStore
from eventy.record import Record, Response, RecordType
from eventy.trace_id.local import local_trace

__all__ = [
    'EventyApp',
]

logger = logging.getLogger(__name__)


class EventyApp:

    def __init__(
        self,
        record_store: RecordStore,
        app_service: Service,
        ext_services: Iterable[Service] = None,
        agents: Iterable[Agent] = None,
        handlers: Iterable[Handler] = None,
        read_batch_size: int = 1,
        read_timeout_ms: Optional[int] = None,
        write_timeout_ms: Optional[int] = None,
        ack_timeout_ms: Optional[int] = None,
        commit_timeout_ms: Optional[int] = None,
    ):
        self._record_store = record_store
        self._app_service = app_service
        self._ext_services = ext_services or []
        self._all_handlers = list()
        if handlers is not None:
            for handler in handlers:
                self._all_handlers.append(handler)
        if agents is not None:
            for agent in agents:
                for handler in agent.handlers:
                    self._all_handlers.append(handler)

        self._read_batch_size = read_batch_size
        self._read_timeout_ms = read_timeout_ms
        self._write_timeout_ms = write_timeout_ms
        self._ack_timeout_ms = ack_timeout_ms
        self._commit_timeout_ms = commit_timeout_ms

        self._write_store = _AppRecordWriteStore(self)

    def _get_service_for_name(self, service_name: Optional[str]):
        if service_name is None:
            return self._app_service
        for service in self._ext_services:
            if service.name == service_name:
                return service
        raise ValueError(f"Service {service_name} not found.")

    def _get_topics_to_register(self) -> list[str]:
        topics: set[str] = set()
        topics.add(self._app_service.request_topic)
        for handler in self._all_handlers:
            logger.debug(f"Found handler {handler}.")
            service = self._get_service_for_name(handler.service_name)
            topics.add(service.topic_for(handler.record_type))
        return list(sorted(topics))

    def _get_handlers_for_record(self, record: Record, guarantee: Guarantee) -> list[Handler]:
        handlers: list[Handler] = list()
        for handler in self._all_handlers:
            if (
                handler.record_type == record.type
                and handler.record_name == record.name
                and handler.delivery_guarantee == guarantee
                and self._get_service_for_name(handler.service_name).namespace == record.namespace
            ):
                if isinstance(record, Response):
                    if record.destination == self._app_service.namespace:
                        handlers.append(handler)
                else:
                    handlers.append(handler)
        return handlers

    def _get_topic_for_record(self, record: Record) -> str:
        if record.type == RecordType.EVENT or record.type == RecordType.RESPONSE:
            return self._app_service.topic_for(record.type)
        for service in self._ext_services:
            if service.namespace == record.namespace:
                return service.topic_for(record.type)
        raise ValueError(f"No topic for record {record}.")

    @property
    def write_store(self) -> RecordWriteStore:
        return self._write_store

    def run(self) -> None:
        for topic in self._get_topics_to_register():
            logger.debug(f"Will register to topic: {topic}.")
            self._record_store.register_topic(topic, Cursor.ACKNOWLEDGED)

        while True:
            records = list(
                self._record_store.read(
                    max_count=self._read_batch_size,
                    timeout_ms=self._read_timeout_ms,
                    auto_ack=False,
                )
            )
            if records:
                logger.debug(f"Received {len(records)} new records.")
            else:
                time.sleep(0.05)
                continue

            # At Least Once
            for record in records:
                with local_trace(correlation_id=record.correlation_id):
                    for handler in self._get_handlers_for_record(record, Guarantee.AT_LEAST_ONCE):
                        for output_record in handler.handle_record(record):
                            self._record_store.write(
                                output_record,
                                self._get_topic_for_record(output_record),
                                self._write_timeout_ms,
                            )

            # Exactly Once
            self._record_store.start_transaction()
            self._record_store.ack(self._ack_timeout_ms)
            for record in records:
                with local_trace(correlation_id=record.correlation_id):
                    for handler in self._get_handlers_for_record(record, Guarantee.EXACTLY_ONCE):
                        for output_record in handler.handle_record(record):
                            self._record_store.write(
                                output_record,
                                self._get_topic_for_record(output_record),
                                self._write_timeout_ms,
                            )
            self._record_store.commit(self._commit_timeout_ms)

            # At Most Once
            for record in records:
                with local_trace(correlation_id=record.correlation_id):
                    for handler in self._get_handlers_for_record(record, Guarantee.AT_MOST_ONCE):
                        for output_record in handler.handle_record(record):
                            self._record_store.write(
                                output_record,
                                self._get_topic_for_record(output_record),
                                self._write_timeout_ms,
                            )


class _AppRecordWriteStore(RecordWriteStore):
    def __init__(self, app: EventyApp):
        self._app = app

    def write(self, record: Record) -> None:
        with local_trace(correlation_id=record.correlation_id):
            self._app._record_store.write(
                record,
                self._app._get_topic_for_record(record),
                self._app._write_timeout_ms,
            )



# Copyright (c) Qotto, 2021

"""
AioHttp integration utilities

Utility functions to integrate the eventy protocol in apps using aiohttp.

Usage::

    import eventy.integration.aiohttp
"""
import logging
from typing import Callable, Awaitable

import aiohttp as vendor_aiohttp
import aiohttp.client as vendor_aiohttp_client
import aiohttp.web as vendor_aiohttp_web
import aiohttp.web_middlewares as vendor_aiohttp_web_middlewares
import aiohttp.web_request as vendor_aiohttp_web_request
import aiohttp.web_response as vendor_aiohttp_web_response

import eventy.config.aiohttp
from eventy.trace_id import correlation_id_var, request_id_var
from eventy.trace_id.generator import gen_trace_id

__all__ = [
    'application',
    'client_session',
]

logger = logging.getLogger(__name__)


def client_session(*args, **kwargs) -> vendor_aiohttp_client.ClientSession:
    """
    AioHttp client session propagating correlation_id
    """

    async def provide_correlation_id_header(
        session, trace_config_ctx, params
    ):
        # headers is a mutable param
        # https://docs.aiohttp.org/en/stable/tracing_reference.html#tracerequeststartparams
        headers: dict = params.headers
        headers.update({'x-correlation-id': correlation_id_var.get()})

    correlation_id_trace_config = vendor_aiohttp.TraceConfig()
    correlation_id_trace_config.on_request_start.append(provide_correlation_id_header)
    trace_configs = kwargs.get('trace_configs', [])
    trace_configs.append(correlation_id_trace_config)
    kwargs['trace_configs'] = trace_configs
    return vendor_aiohttp_client.ClientSession(*args, **kwargs)


@vendor_aiohttp_web_middlewares.middleware
async def _trace_middleware(
    request: vendor_aiohttp_web_request.Request,
    handler: Callable[[vendor_aiohttp_web_request.Request], Awaitable[vendor_aiohttp_web_response.StreamResponse]]
) -> vendor_aiohttp_web_response.StreamResponse:
    # generate trace ids
    request_id_var.set(gen_trace_id(request.path))
    correlation_id = None
    for header_name, header_value in request.headers.items():
        if header_name.lower() == 'x-correlation-id':
            correlation_id = header_value
            break
    correlation_id_var.set(correlation_id or gen_trace_id(request.path))

    # decide if access should be logged (skip healthchecks)
    log_access: bool = not (
        request.method == 'GET'
        and request.path == eventy.config.aiohttp.AIOHTTP_WEB_ACCESS_HEALTH_ROUTE
        and eventy.config.aiohttp.AIOHTTP_WEB_ACCESS_DISABLE_HEALTH_LOGGING
    )

    # log request
    if log_access:
        logger.info(f'Request: {request.method} {request.path} ({request.content_length or 0})')

    # handle request
    response = await handler(request)

    # log response
    if log_access:
        logger.info(f'Response: {request.method} {request.path} {response.status} ({response.content_length})')

    return response


def application(*args, **kwargs) -> vendor_aiohttp_web.Application:
    app = vendor_aiohttp_web.Application(*args, **kwargs)
    app.middlewares.append(_trace_middleware)
    return app
